"""
Author: Thomas M. Lloyd
Master's thesis Spring 2021
"""

import pandas as pd
import gurobipy as gp
from gurobipy import *
import numpy as np
from gurobipy import GRB
import plotly.express as px
import matplotlib.pyplot as plt

#Dataframes
dfV = pd.read_excel("Vessel_Info1.xlsx", sheet_name='Vessels', index_col='Vessel_name')
dfT = pd.read_excel("Weather_Info.xlsx", sheet_name='Duration2')
dfVC = pd.read_excel("Vessel_Info1.xlsx", sheet_name='Act_req', index_col=0)
dfVN = pd.read_excel("Vessel_Info1.xlsx", sheet_name='Dur_Vessel', index_col=0)
dfST = pd.read_excel("Weather_Info.xlsx", sheet_name='Sail_vessel2', index_col=0)
dfWL = pd.read_excel("Vessel_Info1.xlsx", sheet_name='Weather_lim', index_col=0)
dfWh = pd.read_excel("Vessel_Info1.xlsx", sheet_name='Weather_Hs', index_col = 0)
data_time = pd.read_excel("Activity_Info.xlsx", sheet_name='Act_time2', index_col=0)

# Developing schedule

dfT['end'] = dfT['start'] + dfT['duration']
fig = px.timeline(dfT, x_start="start", x_end="end", y="Activity_id")
fig.update_yaxes(autorange="reversed") 

fig.layout.xaxis.type = 'linear'
fig.update_layout(
    title="Schedule for one turbine with Hs = 2.5m",
    xaxis_title="Number of days",
    yaxis_title="Operations",
)
fig.data[0].x = dfT.duration.tolist()
f = fig.full_figure_for_development(warn=False)
fig.show()

##Sets##
activities = [0, 1, 2, 3, 4, 5, 6] # indexed by i
vessels = [v for v in range(dfV.shape[0])]
time_periods = np.arange(0,4,1)
noTurbines = 1
length_period = 7
bigM = 10000


##Parameters##
op_req = dfVC.values.tolist()
operation_requirement = np.asarray(op_req)
w_lim = dfWL.values.tolist()
Weather_lim = np.asarray(w_lim)
t_op = dfVN.values.tolist()
time_operation = np.asarray(t_op)
s_op = dfST.values.tolist()
Sail_vessel = np.asarray(s_op)
Vessel_Time = data_time.values.tolist()
Vessel_TimePeriod =np.asarray(Vessel_Time)

# Weather parameter
Hs = dfWh.values.tolist()
Weather_Hs = np.asarray(Hs)

a = []

for i in range(0,3):
    a.append(Weather_lim - Weather_Hs[:,i])
Period0 = a[0]
Period1 = a[1]
Period2 = a[2]

# Costs
cost_TC = dfV.Day_rate.tolist()
cost_mobilization = dfV.Mob_cost.tolist()

# ------ MODEL-----------------------
m = Model('vessel_configuration')

# Variables
 #------ VARIABLES --------------------
vessel_job = m.addVars(vessels, time_periods, activities, vtype=GRB.BINARY, name="vessel_job")
vessel_selected = m.addVars(vessels, activities, vtype=GRB.BINARY, name="vessel_selected")

# ------ OBJECTIVE FUNCTION -----------
m.setObjective(sum(vessel_selected[v, i] * cost_mobilization[v] for v in vessels for i in activities) +
               sum(vessel_job[v, t, i] * cost_TC[v] * (time_operation[v,i]+Sail_vessel[v,i])
                   for v in vessels for i in activities for t in time_periods), GRB.MINIMIZE)


##Constraints##

# Operation constraint
for v in range(len(vessels)):
    for t in range(len(time_periods)):
        m.addConstr(
           (sum(vessel_job[v, t, i] for i in range(len(activities))) <= 1), "taskOne")
        
# Flow-conservation constraint
for v in vessels:
    for i in activities:
        m.addConstr(
            (sum(vessel_job[v, t, i] for t in time_periods) <= bigM * vessel_selected[v, i]), "selection_requirement")
        
# Time constraint       
for t in time_periods:
    for i in activities:
        m.addConstr(
            (sum(vessel_job[v, t, i] for v in vessels) == Vessel_TimePeriod[i, t]), "task_completion")
          
# Weather constraint    
for v in range(len(vessels)):
    for i in range(len(activities)):
        m.addConstr(
            (sum(vessel_job[v, t, i] * Period2[v,i] for t in range(len(time_periods))) >= 0), "Weather")

# Requirement constraint        
for v in range(len(vessels)):
    for i in range(len(activities)):
        m.addConstr(
            (sum(vessel_job[v, t, i] for t in range(len(time_periods))) <= bigM* operation_requirement[v,i]), "operation_requirement")
        
        
## SOLUTION###

# Compute optimal solution
m.optimize()
m.write("out.sol")
obj = m.getObjective()
best_objective = obj.getValue()
runtime = m.Runtime
MIPGap = m.MIPGap
print(MIPGap)
print(runtime)

# ------- SOLUTION PRINT ------------------
if m.status == GRB.Status.OPTIMAL:
    solution = m.getAttr('x', vessel_job)
    solution2 = m.getAttr('x', vessel_selected)
    A = {}
    B = {}
    C = {}
    for t in time_periods:
        for i in activities:
            for v in vessels:
                if solution[v, t, i] > 0:
                    time = (time_operation[v, i] + Sail_vessel[v,i])
                    cost_charter = cost_TC[v]
                    tot1_cost = cost_TC[v]*(time_operation[v, i] + Sail_vessel[v,i])*solution[v,t,i]
                    print('\nOptimal vessel for task %s in time period %s is: %s' % (i, t, v))
                    if t not in A.keys():
                        A.update({t: [i, v, time, cost_charter]})
                        B.update({t: [v]})
                        C.update({t: [tot1_cost]})
                    else:
                        A[t].append(i)
                        A[t].append(v)
                        A[t].append(time)
                        A[t].append(cost_charter)
                        B[t].append(v)
                        C[t].append(tot1_cost)
                        
# Calculations Analysis
information = {}
utilization_cost = []
utilization_time = []
total_time = 0
total_cost = 0  # Total cost for a time period
actual_total_cost = 0  # Actual cost for time period
sum_cost = 0
time0 = 0
time3 = 0
no0 = 0
no3 = 0

for i in A:
    value = int(len(A[i])/4)
    for j in range(value):
        step = j*4
        if A[i][step] == 0:
            time0 = noTurbines*A[i][2+step]
            no0 = no0 + 1
        elif A[i][step] == 3:
            time3 = noTurbines*A[i][2+step]
            no3 = no3 + 1

timeAvg0 = time0/no0
timeAvg3 = time3/no3

for i in A:
    value = int(len(A[i])/4)
    for j in range(value):
        step = j*4
        if A[i][step] == 0:
            time = timeAvg0
        elif A[i][step] == 3:
            time = timeAvg3
        else:
            time = A[i][2+step]
        total_time = total_time + time
        vessel_cost = cost_TC[A[i][1+step]]
        total_cost = total_cost + vessel_cost*time
        actual_total_cost = actual_total_cost + vessel_cost*length_period
    sum_cost = sum_cost + total_cost
    utility_time = total_time/((len(A[i])/4)*length_period)
    utility_cost = total_cost/actual_total_cost
    information[i] = utility_time, utility_cost
    utilization_time.append(utility_time)
    utilization_cost.append(utility_cost)
    total_time = 0
    total_cost = 0
    actual_total_cost = 0
    

# Averages
avg_utilization_time = sum(utilization_time)/len(utilization_time)
avg_utilization_cost = sum(utilization_cost)/len(utilization_cost)

a = [avg_utilization_time]*len(time_periods)


# Average cost
avg_cost = m.objVal/noTurbines


# Plots the utility
fig = plt.figure()
ax = fig.add_subplot(111)
y_pos = time_periods
ax.bar(y_pos, utilization_time, color='silver', width=0.6, linewidth=3, label='Utilization Time Period')
ax.plot(y_pos, a, color='firebrick', linewidth=3, label='Average Utilization Time')
ax.plot(y_pos, utilization_time, color='lightblue', linewidth=3, label='Utilization Time')

values = np.arange(0, len(time_periods), 1)

# Add title and axis names
plt.title('Time Utilization Installation Two Turbines with Hs = 2.5m')
plt.xlabel('Time Periods')
plt.xticks(np.arange(min(time_periods), max(time_periods)+1, 1.0))
plt.ylabel('Utilization')
plt.legend(loc='lower left')
plt.xticks(values, ['%d' % val for val in values])
plt.savefig('Utility Installation Two Turbines with Hs = 2.5m.png')
plt.show()